{% extends "auth/page.html.j2" %}

{% block main %}

<div id="logout-main" class="container">
  <p>
    Please log in via your learning management system (Moodle, Canvas, etc.)!
  </p>
</div>

{% endblock %}