# Default Grader Convert Configuration File

# -------------------------------------------------
# ClearSolutions Configuration
# -------------------------------------------------

# `code_stub` replaces the content of solution cells with a language-specific code snippet.
# Instructors can override these placeholders with the code snippets of their choice.
# c.ClearSolutions.code_stub = {
    # 'python': "# YOUR CODE HERE\nraise NotImplementedError()",  # Placeholder for Python solution cells
    # 'matlab': "% YOUR CODE HERE\nerror('No Answer Given!')",  # Placeholder for MATLAB solution cells
    # 'octave': "% YOUR CODE HERE\nerror('No Answer Given!')",  # Placeholder for Octave solution cells
    # ... (More languages)
# }

# -------------------------------------------------
# ClearAlwaysHiddenTests Configuration
# -------------------------------------------------

# Delimiters for always hidden utilities.
# Blocks of code between these delimiters will always be hidden in the notebook.
# c.ClearAlwaysHiddenTests.begin_util_delimeter = "BEGIN ALWAYS HIDDEN TESTS"  # Start delimiter
# c.ClearAlwaysHiddenTests.end_util_delimeter = "END ALWAYS HIDDEN TESTS"  # End delimiter

# -------------------------------------------------
# ClearHiddenTests Configuration
# -------------------------------------------------

# Delimiters for hidden tests.
# Blocks of code between these delimiters will be hidden in the notebook.
# c.ClearHiddenTests.begin_test_delimeter = "BEGIN HIDDEN TESTS"  # Start delimiter
# c.ClearHiddenTests.end_test_delimeter = "END HIDDEN TESTS"  # End delimiter

# -------------------------------------------------
# ClearMarkScheme Configuration
# -------------------------------------------------

# Delimiters for the mark scheme.
# Blocks of text between these delimiters describe the marking scheme.
# c.ClearMarkScheme.begin_mark_scheme_delimeter = "BEGIN MARK SCHEME"  # Start delimiter
# c.ClearMarkScheme.end_mark_scheme_delimeter = "END MARK SCHEME"  # End delimiter

# -------------------------------------------------
# IncludeHeaderFooter Configuration
# -------------------------------------------------

# Header and Footer files to be included at the top and bottom of each notebook.
# c.IncludeHeaderFooter.header = "../../source/<assignment_id>/header.ipynb"  # Header notebook file
# c.IncludeHeaderFooter.footer = "../../source/<assignment_id>/footer.ipynb"  # Footer notebook file

# -------------------------------------------------
# LimitOutput Configuration
# -------------------------------------------------

# Limit the number of lines and traceback lines in the output cells.
# c.LimitOutput.max_lines = 1000  # Max number of lines in output
# c.LimitOutput.max_traceback = 100  # Max number of traceback lines

# -------------------------------------------------
# LockCells Configuration
# -------------------------------------------------

# Options for locking cells in the notebook to prevent editing.
# c.LockCells.lock_solution_cells = True  # Lock solution cells
# c.LockCells.lock_grade_cells = True  # Lock grade cells
# c.LockCells.lock_readonly_cells = True  # Lock readonly cells
# c.LockCells.lock_all_cells = False  # Lock all cells in the notebook (overrides above settings)

# -------------------------------------------------
# END OF CONFIGURATION FILE
# -------------------------------------------------
